import logging
from typing import Iterator, Dict

from algoliasearch.search_client import SearchClient

from nintendeals.constants import SWITCH, N3DS

APP_ID = "U3B6GR4UA3"
API_KEY = "9a20c93440cf63cf1a7008d75f7438bf"

INDEX_NAME = "noa_aem_game_en_us_title_asc"
INDEX = None


PLATFORM_CODES = {
    SWITCH: "7001",
    N3DS: "5001",
}

log = logging.getLogger(__name__)


def _search_index(query, **options):
    global INDEX

    if not INDEX:
        client = SearchClient.create(APP_ID, API_KEY)
        INDEX = client.init_index(INDEX_NAME)

    log.info("Searching index for %s", query)

    response = INDEX.search(query, request_options=options)
    return response.get('hits', [])


def _search_by_nsuid(platform: str) -> Iterator[Dict]:
    empty_pages = 0

    platform_code = PLATFORM_CODES[platform]

    options = {
        "allowTyposOnNumericTokens": False,
        "queryType": "prefixAll",
        "restrictSearchableAttributes": ["nsuid"],
        "facetFilters": [
            f"platform:{platform}"
        ],
        "hitsPerPage": 500,
    }

    current = -1

    while True:
        current += 1
        query = f"{platform_code}{current:07}"
        games = _search_index(query, **options)

        if not games:
            empty_pages += 1

        if empty_pages == 5:
            break

        yield from games


def _search_by_query(platform: str, query: str) -> Iterator[Dict]:
    hits_per_page = 50

    options = {
        "facetFilters": [
            f"platform:{platform}"
        ],
        "hitsPerPage": hits_per_page,
    }

    page = -1
    while True:
        page += 1
        options["page"] = page

        games = _search_index(query, **options)

        yield from games

        if len(games) < hits_per_page:
            break


def search_games(platform: str, **kwargs) -> Iterator[Dict]:
    query = kwargs.get("query")

    if query:
        yield from _search_by_query(platform, query)
    else:
        yield from _search_by_nsuid(platform)


def find_by_nsuid(nsuid: str) -> str:
    hits = _search_index(
        nsuid,
        attributesToRetrieve=["title", "nsuid", "slug"],
        restrictSearchableAttributes=['nsuid'],
    )

    return hits[0]["slug"]



from typing import Optional

from nintendeals.classes.games import Game


class SwitchGame(Game):

    platform = "Nintendo Switch"

    def __init__(
        self,
        region: str,
        title: str,
        nsuid: str = None,
        product_code: str = None,
    ):
        super().__init__(
            region=region,
            title=title,
            nsuid=nsuid,
            product_code=product_code
        )

        self.game_vouchers: bool = None
        self.local_multiplayer: bool = None
        self.nso_required: bool = None
        self.save_data_cloud: bool = None
        self.voice_chat: str = None

    @property
    def unique_id(self) -> Optional[str]:
        if not self.product_code:
            return None

        return self.product_code[-5:-1]



from nintendeals.constants import REGIONS


class InvalidAlpha2Code(ValueError):
    def __init__(self, country_code: str):
        super().__init__(
            f"{country_code} is not a valid alpha-2 code. Visit"
            f" https://iso.org/obp/ui/#search/code for a list of valid codes."
        )


class InvalidNsuidFormat(ValueError):
    def __init__(self, nsuid: str):
        super().__init__(
            f"{nsuid} is not a valid nsuid, a 14 digit string was expected."
        )


class InvalidRegion(ValueError):
    def __init__(self, region: str):
        super().__init__(
            f"{region} is not a valid Nintendo region, the valid regions"
            f" are: {', '.join(REGIONS)}."
        )


class InvalidTitle(TypeError):
    def __init__(self):
        super().__init__("The title cannot be an empty string or None")



from datetime import datetime
from typing import List, Optional

from nintendeals import validate
from nintendeals.api.prices import get_price
from nintendeals.classes.prices import Price


class Game:

    platform = None

    @validate.title
    @validate.region
    @validate.nsuid(nullable=True)
    def __init__(
        self,
        region: str,
        title: str,
        nsuid: str = None,
        product_code: str = None,
    ):
        self.region: str = region
        self.title: str = title
        self.nsuid: str = nsuid
        self.product_code: str = product_code

        self.slug: str = None

        self.description: str = None
        self.developer: str = None
        self.genres: List[str] = []
        self.languages: List[str] = []
        self.megabytes: int = None
        self.players: int = 0
        self.publisher: str = None
        self.release_date: datetime = None

        # Features
        self.amiibo: bool = None
        self.demo: bool = None
        self.dlc: bool = None
        self.free_to_play: bool = None
        self.iaps: bool = None

    @property
    def unique_id(self) -> Optional[str]:
        raise NotImplementedError()

    @validate.country
    def price(self, *, country: str) -> Price:
        """
            Using the price API it will retrieve the pricing
        of the game in the given country. It will only work if
        the country is is under the region of the game.

        Parameters
        ----------
        country: str
            Valid alpha-2 code of the country.

        Returns
        -------
        nintendeals.classes.prices.Price
            Pricing of this game in the given country.

        Raises
        -------
        nintendeals.exceptions.InvalidAlpha2Code
            The `country` wasn't a valid alpha-2 code.
        """
        return get_price(country=country, game=self)

    @validate.country
    def url(self, *, country: str, lang: str = "en") -> Optional[str]:
        """
            Given a valid `country` code and an optional language it will
        provide a url (using the game's nsuid) that will redirect to the
        eShop page for this game.

        Parameters
        ----------
        country: str
            Valid alpha-2 code of the country.
        lang: str = "en"
            Valid iso-code for language.

        Returns
        -------
        str
            URL for the eShop of the game.

        Raises
        -------
        nintendeals.exceptions.InvalidAlpha2Code
            The `country` wasn't a valid alpha-2 code.
        """
        if not self.nsuid:
            return None

        return f"https://ec.nintendo.com/{country}/{lang}/titles/{self.nsuid}"

    def __repr__(self):
        return self.title



import logging
from datetime import datetime
from typing import Iterator, Type, Union

import requests

from nintendeals.classes import N3dsGame, SwitchGame
from nintendeals.constants import EU, N3DS, SWITCH

log = logging.getLogger(__name__)


def _list_games(
    game_class: Type,
    system_name: str,
    **kwargs
) -> Iterator[Union[N3dsGame, SwitchGame]]:
    query = kwargs.get("query", "*")
    nsuid = kwargs.get("nsuid")

    if not nsuid:
        fq = f"type:GAME AND system_names_txt:\"{system_name}\""
    else:
        fq = f"type:GAME AND nsuid_txt:\"{nsuid}\""

    rows = 200
    start = -rows

    while True:
        start += rows

        params = {
            "q": query,
            "wt": "json",
            "sort": "title asc",
            "start": start,
            "rows": rows,
            "fq": fq
        }

        response = requests.get(
            url='https://search.nintendo-europe.com/en/select',
            params=params
        )

        if response.status_code != 200:
            break

        json = response.json().get('response').get('docs', [])

        if not len(json):
            break

        for data in json:
            product_codes = [
                pc for pc in data.get("product_code_txt", [])
                if "-" not in pc
            ] or [None]

            nsuids = list(reversed(sorted(data.get("nsuid_txt", [None]))))

            game = game_class(
                region=EU,
                title=data["title_extras_txt"][0],
                nsuid=nsuids[0],
                product_code=product_codes[0],
            )

            game.slug = data.get("url")

            game.description = data.get("excerpt")
            game.developer = data.get("developer")
            game.publisher = data.get("publisher")
            game.genres = data.get("pretty_game_categories_txt", [])
            game.players = data.get("players_to")
            game.languages = list(sorted(map(
                lambda lang: lang.title(),
                data.get("language_availability", [""])[0].split(",")
            )))

            try:
                game.release_date = datetime.strptime(
                    data["dates_released_dts"][0].split("T")[0],
                    '%Y-%m-%d'
                )
            except (ValueError, TypeError):
                game.release_date = None

            if "datasize_readable_txt" in data:
                value, unit = data.get("datasize_readable_txt", [""])[0].split()

                if unit.lower() == "blocks":
                    value = int(value) // 8
            else:
                value = None

            game.megabytes = value

            # Features
            game.amiibo = data.get("near_field_comm_b", False)
            game.demo = data.get("demo_availability", False)
            game.dlc = data.get("dlc_shown_b", False)
            game.free_to_play = data.get("price_regular_f") == 0.0

            if game.platform == N3DS:
                game.download_play = data.get("download_play", False)
                game.motion_control = data.get("motion_control_3ds", False)
                game.spot_pass = data.get("spot_pass", False)
                game.street_pass = data.get("street_pass", False)
                game.virtual_console = '3ds_virtualconsole' in data.get("system_type", [""])[0]

            if game.platform == SWITCH:
                game.local_multiplayer = data.get("local_play", False)
                game.nso_required = data.get("paid_subscription_required_b", False)
                game.save_data_cloud = data.get("cloud_saves_b", False)
                game.game_vouchers = data.get("switch_game_voucher_b", False)
                game.voice_chat = data.get("voice_chat_b", False)

            yield game


def list_3ds_games(**kwargs) -> Iterator[N3dsGame]:
    """
        List all the 3DS games in Nintendo of Europe. The following subset
    of data will be available for each game.

    Game data
    ---------
        * platform: str ["Nintendo 3DS"]
        * region: str ["EU"]
        * title: str
        * nsuid: str (optional)
        * product_code: str (optional)

        * slug: str

        * amiibo: bool
        * demo: bool
        * description: str
        * developer: str
        * dlc: bool
        * download_play: bool
        * free_to_play: bool
        * genres: List[str]
        * languages: List[str]
        * megabytes: int
        * motion_control: bool
        * players: int
        * publisher: str
        * release_date: datetime
        * spot_pass: bool
        * street_pass: bool
        * virtual_console: bool

    Yields
    -------
    nintendeals.classes.N3dsGame:
        3DS game from Nintendo of Europe.
    """
    log.info("Fetching list of Nintendo 3DS games")

    yield from _list_games(
        game_class=N3dsGame,
        system_name="3ds",
        **kwargs
    )


def list_switch_games(**kwargs) -> Iterator[SwitchGame]:
    """
        List all the Switch games in Nintendo of Europe. The following subset
    of data will be available for each game.

    Game data
    ---------
        * platform: str ["Nintendo Switch"]
        * region: str ["EU"]
        * title: str
        * nsuid: str (optional)
        * product_code: str (optional)

        * slug: str

        * amiibo: bool
        * demo: bool
        * developer: str
        * dlc: bool
        * free_to_play: bool
        * game_vouchers: bool
        * genres: List[str]
        * languages: List[str]
        * local_multiplayer: bool
        * nso_required: bool
        * players: int
        * publisher: str
        * release_date: datetime
        * save_data_cloud: bool
        * voice_chat: bool

    Yields
    -------
    nintendeals.classes.SwitchGame:
        Switch game from Nintendo of Europe.
    """
    log.info("Fetching list of Nintendo Switch games")

    yield from _list_games(
        game_class=SwitchGame,
        system_name="Switch",
        **kwargs
    )



import re

from pycountry import countries

from nintendeals import exceptions
from nintendeals.constants import REGIONS

NSUID_REGEX = re.compile(r"[5-7]00\d{11}")


def _validate_nsuid(nsuid_: str, nullable: bool = False):
    if nullable and nsuid_ is None:
        return

    if not isinstance(nsuid_, str):
        raise exceptions.InvalidNsuidFormat(nsuid_)

    if not NSUID_REGEX.match(nsuid_):
        raise exceptions.InvalidNsuidFormat(nsuid_)


def _validate_game(game: "Game", nullable: bool = False):
    if nullable and game is None:
        return

    _validate_nsuid(game.nsuid)


def country(func):
    """
        Validates that the parameter `country` of the decorated function is
    a valid alpha-2 country code.
        Full list of valid codes at https://www.iso.org/obp/ui/#search/code.

    Examples
    -------
    >>> country="CZ"      # ✅️
    >>> country="XX"      # ❌️
    >>> country="ARG"     # ❌️
    >>> country=123       # ❌️

    Raises
    -------
    nintendeals.exceptions.InvalidAlpha2Code
        The `code` wasn't a valid alpha-2 code.
    """
    def wrapper(*args, **kwargs):
        code_ = kwargs.get("country")
        country_ = countries.get(alpha_2=code_)

        if not country_:
            raise exceptions.InvalidAlpha2Code(code_)

        return func(*args, **kwargs)
    return wrapper


def region(func):
    """
        Validates that the parameter `region` of the decorated function is
    a valid Nintendo region.

    Examples
    -------
    >>> region="NA"    # ✅️
    >>> region="LA"    # ❌️
    >>> region="ASIA"  # ❌️

    Raises
    -------
    nintendeals.exceptions.InvalidRegion
        The `region` wasn't valid.
    """
    def wrapper(*args, **kwargs):
        region_ = kwargs.get("region")

        if region_ not in REGIONS:
            raise exceptions.InvalidRegion(region_)

        return func(*args, **kwargs)
    return wrapper


def title(func):
    """
        Validates that the parameter `title` of the decorated function is
    not empty or none.

    Examples
    -------
    >>> title="Super Mario Party"    # ✅️
    >>> title=""                     # ❌️
    >>> title=None                   # ❌️

    Raises
    -------
    nintendeals.exceptions.InvalidTitle
        The `title` wasn't valid.
    """
    def wrapper(*args, **kwargs):
        if not title:
            raise exceptions.InvalidTitle()

        return func(*args, **kwargs)
    return wrapper


def nsuids(nullable: bool = False):
    """
        Validates that all the nsuids in the parameter `nsuids` of the decorated
    function match the format of a valid nsuid.
        A valid nsuid follows the "[5-7]00[0-9]{11}" regular expression.

    Parameters
    ----------
    nullable: bool = False
        If true, an exception won't be raise if any nsuid is None.

    Examples
    -------
    >>> nsuids=["70010000000450"]  # ✅️
    >>> nsuids=["50010000000654"]  # ✅️
    >>> nsuids=["70010000000"]     # ❌️
    >>> nsuids=[50010000000]       # ❌️

    Raises
    -------
    nintendeals.exceptions.InvalidNsuidFormat
        Any of the `nsuids` was either none or had an invalid format.
    """
    def outer(func):
        def inner(*args, **kwargs):
            for nsuid_ in kwargs.get("nsuids", []):
                _validate_nsuid(nsuid_, nullable)

            return func(*args, **kwargs)
        return inner

    if callable(nullable):
        return outer(func=nullable)

    return outer


def nsuid(nullable: bool = False):
    """
        Validates that the parameter `nsuid` of the decorated function matches
    the format of a valid nsuid.
        A valid nsuid follows the "[5-7]00[0-9]{11}" regular expression.

    Parameters
    ----------
    nullable: bool = False
        If true, an exception won't be raise if the nsuid is None.

    Examples
    -------
    >>> nsuids=["70010000000450"]  # ✅️
    >>> nsuids=["50010000000654"]  # ✅️
    >>> nsuids=["70010000000"]     # ❌️
    >>> nsuids=[50010000000]       # ❌️

    Raises
    -------
    nintendeals.exceptions.InvalidNsuidFormat
        The `nsuids` was either none or had an invalid format.
    """
    def outer(func):
        def inner(*args, **kwargs):
            nsuid_ = kwargs.get("nsuid")
            _validate_nsuid(nsuid_, nullable)

            return func(*args, **kwargs)
        return inner

    if callable(nullable):
        return outer(func=nullable)

    return outer


def games(nullable: bool = False):
    """
        Validates that all the games in the parameter `games` of the decorated
    function follow a list of restrictions.

    Parameters
    ----------
    nullable: bool = False
        If true, an exception won't be raise if any game is None.

    Raises
    -------
    nintendeals.exceptions.InvalidNsuidFormat
        Any of the `games` had an nsuid that was either none or
    had an invalid format.
    """
    def outer(func):
        def inner(*args, **kwargs):
            for game_ in kwargs.get("games", []):
                _validate_game(game_, nullable)

            return func(*args, **kwargs)
        return inner

    if callable(nullable):
        return outer(func=nullable)

    return outer


def game(nullable: bool = False):
    """
        Validates that the parameter `game` of the decorated function follows
     a list of restrictions.

    Parameters
    ----------
    nullable: bool = False
        If true, an exception won't be raise if the game is None.

    Raises
    -------
    nintendeals.exceptions.InvalidNsuidFormat
        The `games` had an nsuid that was either none or had an invalid format.
    """
    def outer(func):
        def inner(*args, **kwargs):
            game_ = kwargs.get("game")
            _validate_game(game_, nullable)

            return func(*args, **kwargs)
        return inner

    if callable(nullable):
        return outer(func=nullable)

    return outer



import logging
from datetime import datetime
from typing import Type, Union

import requests
from bs4 import BeautifulSoup

from nintendeals import validate
from nintendeals.classes import N3dsGame, SwitchGame
from nintendeals.classes.games import Game
from nintendeals.constants import EU
from nintendeals.noe.listing import list_3ds_games

log = logging.getLogger(__name__)


def _sibling(soup: BeautifulSoup, string: str, tag: str = "p") -> str:
    p = soup.find(tag, class_="game_info_title", string=string)

    if not p:
        return None

    sib = p.find_next_sibling("p")

    if not sib:
        return None

    return sib.text


def _scrap_switch(nsuid: str) -> Game:
    url = f"https://ec.nintendo.com/GB/en/titles/{nsuid}"
    response = requests.get(url, allow_redirects=True)

    if response.status_code != 200:
        return None

    soup = BeautifulSoup(response.text, features="html.parser")

    scripts = list(filter(
        lambda s: "var nsuids = [" in str(s), soup.find_all('script')
    ))

    if not scripts:
        return None

    script = scripts[0]
    lines = [
        line.strip().replace("\",", "") for line in str(script).split("\n")
        if ':' in line
    ]

    data = {}

    for line in lines:
        split = line.split(": \"")

        if len(split) != 2:
            continue

        data[split[0].replace("\"", "")] = split[1].replace("\"", "")

    game = SwitchGame(
        region=EU,
        title=soup.find("h1").text,
        nsuid=data["nsuid"],
        product_code=data["productCode"],
    )

    game.slug = response.url.replace("https://www.nintendo.co.uk", "")

    game.description = soup.find("div", class_="col-xs-12 content").text.strip()
    game.developer = _sibling(soup, "Developer")
    game.publisher = _sibling(soup, "Publisher")

    # Genres
    game.genres = list(sorted(map(
        lambda g: g.strip(), _sibling(soup, string="Categories").split(",")
    )))

    # Languages
    game.languages = list(sorted(
        _sibling(soup, string="Languages").split(", ")
    ))

    # Players
    try:
        text = _sibling(soup, "Players")
        game.players = max(map(int, text.split(" - ")))
    except (AttributeError, ValueError):
        game.players = 0

    # Release date
    try:
        release_date = data["releaseDate"]
        game.release_date = datetime.strptime(release_date, '%d/%m/%Y')
    except ValueError:
        pass

    # Game size (in MBs)
    rom_size = _sibling(soup, "Download size")
    if rom_size:
        value, unit = rom_size.split(" ")
        value = round(float(value))
    else:
        value = None

    game.megabytes = value

    # Other properties
    features = _sibling(soup, string="Features")

    # Common Features
    game.amiibo = "amiibo" in features
    game.demo = "Demo available" in features
    game.dlc = "Downloadable content" in features
    game.free_to_play = "\"offdeviceProductPrice\": \"0.0\"" in response.text
    game.iaps = "Offers in-game purchases" in response.text

    # Switch Features
    game.local_multiplayer = "Local multiplayer" in features
    game.game_vouchers = None  # unsupported
    game.nso_required = "Paid online membership service" in features
    game.save_data_cloud = "Save Data Cloud" in features
    game.voice_chat = "Voice Chat" in features

    return game


@validate.nsuid
def game_info(*, nsuid: str) -> Union[N3dsGame, SwitchGame, Type[None]]:
    """
        Given a valid nsuid for the EU region, it will retrieve the
    information of the game with that nsuid from Nintendo of Europe.

    Game data
    ---------
        * platform: str ["Nintendo 3DS", "Nintendo Switch"]
        * region: str ["EU"]
        * title: str
        * nsuid: str (optional)
        * product_code: str (optional)

        * slug: str

        * amiibo: bool
        * demo: bool
        * description: str
        * developer: str (optional)
        * dlc: bool
        * free_to_play: bool
        * genres: List[str]
        * iaps: bool
        * languages: List[str]
        * megabytes: int
        * players: int
        * publisher: str
        * release_date: datetime

        # Switch Features
        * local_multiplayer: bool
        * game_vouchers: bool (unsupported)
        * nso_required: bool
        * save_data_cloud: bool
        * voice_chat: bool

        # 3DS Features
        * download_play: bool
        * motion_control: bool
        * spot_pass: bool
        * street_pass: bool
        * virtual_console: bool

    Parameters
    ----------
    nsuid: str
        Valid nsuid of a nintendo game.

    Returns
    -------
    nintendeals.classes.N3DSGame:
        3DS game from Nintendo of Europe.
    nintendeals.classes.SwitchGame:
        Switch game from Nintendo of Europe.
    None:
        No game with the provided nsuid was found on Nintendo of Europe.

    Raises
    -------
    nintendeals.exceptions.InvalidNsuidFormat
        The nsuid was either none or has an invalid format.
    """
    if nsuid.startswith("5"):
        log.info("Fetching info for %s", nsuid)
        games = list(list_3ds_games(nsuid=nsuid))
        return games[0] if games else None

    if nsuid.startswith("7"):
        log.info("Fetching info for %s", nsuid)
        return _scrap_switch(nsuid=nsuid)

    return None



import logging
import re
from datetime import datetime
from typing import Iterator, Type, Union

from nintendeals.classes import N3dsGame, SwitchGame
from nintendeals.classes.games import Game
from nintendeals.constants import NA, N3DS
from nintendeals.noa.external import algolia

log = logging.getLogger(__name__)


def _list_games(
    game_class: Type,
    **kwargs
) -> Iterator[Union[N3dsGame, SwitchGame]]:

    for data in algolia.search_games(platform=game_class.platform, **kwargs):
        game = game_class(
            region=NA,
            title=data["title"],
            nsuid=data.get("nsuid"),
        )

        game.slug = data["slug"]
        game.genres = list(sorted(data.get("categories", [])))

        try:
            release_date = data["releaseDateMask"].split("T")[0]
            game.release_date = datetime.strptime(release_date, '%Y-%m-%d')
        except ValueError:
            pass

        try:
            game.players = int(re.sub(r"[^\d]*", "", data["players"]))
        except ValueError:
            game.players = 0

        game.description = data.get("description")
        game.free_to_play = data.get("msrp") == 0.0
        game.publisher = data.get("publishers", [None])[0]
        game.developer = data.get("developers", [None])[0]

        if game.platform == N3DS:
            game.virtual_console = data.get("virtualConsole", "na") != "na"

        yield game


def list_3ds_games(**kwargs) -> Iterator[Game]:
    """
        List all the 3DS games in Nintendo of America. The following subset
    of data will be available for each game.

    Game data
    ---------
        * platform: str ["Nintendo 3DS"]
        * region: str ["NA"]
        * title: str
        * nsuid: str (optional)
        * product_code: str (unsupported)

        * slug: str

        * description: str
        * developer: str
        * free_to_play: bool
        * genres: List[str]
        * publisher: str
        * release_date: datetime
        * virtual_console: bool

    Yields
    -------
    nintendeals.classes.N3dsGame:
        3DS game from Nintendo of America.
    """
    log.info("Fetching list of Nintendo 3DS games")

    yield from _list_games(N3dsGame, **kwargs)


def list_switch_games(**kwargs) -> Iterator[SwitchGame]:
    """
        List all the Switch games in Nintendo of America. The following subset
    of data will be available for each game.

    Game data
    ---------
        * platform: str ["Nintendo Switch"]
        * region: str ["NA"]
        * title: str
        * nsuid: str (optional)
        * product_code: str (unsupported)

        * slug: str

        * description: str
        * developer: str
        * free_to_play: bool
        * genres: List[str]
        * publisher: str
        * release_date: datetime

    Yields
    -------
    nintendeals.classes.SwitchGame:
        Switch game from Nintendo of America.
    """
    log.info("Fetching list of Nintendo Switch games")

    yield from _list_games(SwitchGame, **kwargs)



import logging
from datetime import datetime
from typing import Dict, Iterator, List, Optional, Tuple

import requests
from dateutil.parser import parse as date_parser

from nintendeals import validate
from nintendeals.classes.prices import Price

log = logging.getLogger(__name__)


def _parse_date(string: str) -> datetime:
    return date_parser(string).replace(tzinfo=None)


@validate.country
@validate.nsuids
def _fetch_prices(
    *,
    country: str,
    nsuids: List[str]
) -> Dict[str, Price]:
    if not 51 > len(nsuids) > 0:
        raise ValueError("The amount of nsuids must between 1 and 50.")

    log.info("Calling prices api with %i nsuids", len(nsuids))

    response = requests.get(
        url="https://api.ec.nintendo.com/v1/price",
        params=dict(
            country=country,
            lang="en",
            ids=",".join(nsuids),
        )
    )

    response.raise_for_status()

    total = 0
    sales = 0
    prices = {}

    for data in response.json().get('prices', []):
        nsuid = str(data["title_id"])
        regular_price = data.get("regular_price")

        if not regular_price:
            prices[nsuid] = None
            continue

        total += 1

        price = Price(
            nsuid=nsuid,
            country=country,
            currency=regular_price["currency"],
            value=float(regular_price["raw_value"]),
        )

        discount_price = data.get("discount_price")

        if discount_price:
            price.sale_value = float(discount_price["raw_value"])
            price.sale_start = _parse_date(discount_price['start_datetime'])
            price.sale_end = _parse_date(discount_price['end_datetime'])

            sales += 1

        prices[price.nsuid] = price

    log.info("Found %i prices and %i sales", total, sales)
    return prices


@validate.country
@validate.games
def get_prices(
    *,
    country: str,
    games: List["Game"]
) -> Iterator[Tuple[str, Price]]:
    """
        Given a valid `country` code and a list of `games` (each with a nsuid)
    it will retrieve the current pricing of those games for that country.

    Parameters
    ----------
    country: str
        Valid alpha-2 code of the country.
    games: List[nintendeals.classes.games.Game]
        Games to get their pricing.

    Yields
    -------
    Tuple[str, nintendeals.classes.prices.Price]
        Dictionary containing the nsuids as keys and each pricing as value.

    Raises
    -------
    nintendeals.exceptions.InvalidAlpha2Code
        The `country` wasn't a valid alpha-2 code.
    nintendeals.exceptions.InvalidNsuidFormat
        Any of the `games` had an nsuid that is either none or
    has an invalid format.
    """
    log.info("Fetching prices for %i games in %s", len(games), country)

    prices = {}
    chunk = []

    for game in games:
        chunk.append(game)

        if len(chunk) == 50:
            nsuids = [game.nsuid for game in chunk]
            fetched = _fetch_prices(country=country, nsuids=nsuids)
            prices.update(fetched)

            chunk = []

    if chunk:
        nsuids = [game.nsuid for game in chunk if game.nsuid]
        fetched = _fetch_prices(country=country, nsuids=nsuids)
        prices.update(fetched)

    yield from prices.items()


@validate.country
@validate.game
def get_price(
    *,
    country: str,
    game: "Game"
) -> Optional[Price]:
    """
        Given a valid `country` code and a `game` (with an nsuid) it will
    retrieve the current pricing of that game for that country.

    Parameters
    ----------
    country: str
        Valid alpha-2 code of the country.
    game: nintendeals.classes.games.Game
        Game to get its pricing.

    Returns
    -------
    nintendeals.classes.prices.Price
        Pricing of the game in the indicated country.

    Raises
    -------
    nintendeals.exceptions.InvalidAlpha2Code
        The `country` wasn't a valid alpha-2 code.
    nintendeals.exceptions.InvalidNsuidFormat
        The nsuid was either none or has an invalid format.
    """
    log.info("Fetching price for %s games in %s", game.title, country)

    fetched = _fetch_prices(country=country, nsuids=[game.nsuid])
    return fetched.get(game.nsuid)



import logging
from datetime import datetime
from functools import lru_cache
from typing import Iterator, List, Union, Type

import requests
import xmltodict

from nintendeals.classes import N3dsGame, SwitchGame
from nintendeals.constants import JP

log = logging.getLogger(__name__)


@lru_cache()
def _list_games(
    game_class: Type,
    filename: str
) -> List[Union[N3dsGame, SwitchGame]]:
    url = f"https://www.nintendo.co.jp/data/software/xml/{filename}.xml"
    response = requests.get(url)

    if response.status_code != 200:
        return []

    xml = xmltodict.parse(response.text)['TitleInfoList']['TitleInfo']

    games = []

    for data in xml:
        game = game_class(
            region=JP,
            title=data["TitleName"],
            nsuid=data["LinkURL"].split("/")[-1],
            product_code=data["InitialCode"],
        )

        game.developer = data.get("MakerName")
        game.free_to_play = data.get("Price") == "無料"

        try:
            game.release_date = datetime.strptime(
                data.get('SalesDate'),
                '%Y.%m.%d'
            )
        except (ValueError, TypeError):
            game.release_date = None

        games.append(game)

    return games


def list_3ds_games() -> Iterator[N3dsGame]:
    """
        List all the 3DS games in Nintendo of Japan. The following subset
    of data will be available for each game.

    Game data
    ---------
        * platform: str ["Nintendo 3DS"]
        * region: str ["JP"]
        * title: str
        * nsuid: str
        * product_code: str

        * developer: str
        * free_to_play: bool
        * release_date: datetime (optional)

    Yields
    -------
    nintendeals.classes.N3dsGame:
        3DS game from Nintendo of Japan.
    """
    log.info("Fetching list of Nintendo 3DS games")

    yield from _list_games(N3dsGame, filename="3ds_pkg_dl")


def list_switch_games() -> Iterator[SwitchGame]:
    """
        List all the Switch games in Nintendo of Japan. The following subset
    of data will be available for each game.

    Game data
    ---------
        * platform: str ["Nintendo Switch"]
        * region: str ["JP"]
        * title: str
        * nsuid: str
        * product_code: str

        * developer: str
        * free_to_play: bool
        * release_date: datetime (optional)

    Yields
    -------
    nintendeals.classes.SwitchGame:
        Switch game from Nintendo of Japan.
    """
    log.info("Fetching list of Nintendo Switch games")

    yield from _list_games(SwitchGame, filename="switch")



from datetime import datetime


def search_filter(
    game: "Game",
    title: str = None,
    released_at: datetime = None,
    released_after: datetime = None,
    released_before: datetime = None,
) -> bool:
    if title and title not in game.title:
        return False

    release_date = game.release_date

    if any([released_at, released_after, released_before]) \
            and not release_date:
        return False

    if released_at and release_date != released_at:
        return False

    if released_after and release_date < released_after:
        return False

    if released_before and release_date > released_before:
        return False

    return True



import logging
import re
from datetime import datetime
from typing import Type, Union
from urllib import parse

import requests
from bs4 import BeautifulSoup

from nintendeals import validate
from nintendeals.classes import N3dsGame, SwitchGame
from nintendeals.classes.games import Game
from nintendeals.constants import NA, N3DS, SWITCH
from nintendeals.noa.external import algolia

log = logging.getLogger(__name__)


def _unquote(string: str) -> str:
    return parse.unquote(
        string
        .replace("\\u00", "%")
        .replace("\\x27", "'")
        .replace("\\/", "/")
        .replace(" : ", ": ")
        .strip()
    )


def _aria_label(soup, label, tag="a"):
    tag = soup.find(tag, {"aria-label": label})
    return tag and _unquote(tag.text.strip())


def _class(soup, cl, tag="dd"):
    tag = soup.find(tag, {"class": cl})
    return tag and _unquote(tag.text.strip())


def _itemprop(soup, prop, tag="dd"):
    tag = soup.find(tag, {"itemprop": prop})
    return tag and _unquote(tag.text.strip())


def _scrap(
    game_class: Type,
    slug: str
) -> Union[N3dsGame, SwitchGame]:
    url = f"https://www.nintendo.com/games/detail/{slug}"
    response = requests.get(url, allow_redirects=True)
    soup = BeautifulSoup(response.text, features="html.parser")

    scripts = list(filter(lambda s: "window.game" in str(s), soup.find_all('script')))

    if not scripts:
        return None

    script = scripts[0]
    lines = [line.strip().replace("\",", "") for line in str(script).split("\n") if ':' in line]
    data = dict(map(lambda line: line.split(': "'), lines))

    game = game_class(
        region=NA,
        title=_unquote(data["title"]),
        nsuid=data["nsuid"],
        product_code=data["productCode"],
    )

    game.slug = slug

    game.description = _itemprop(soup, "description", tag="div")
    game.developer = _itemprop(soup, "manufacturer") or None
    game.publisher = _unquote(data["publisher"]) or None

    # Genres
    game.genres = list(sorted([
        genre for genre in _unquote(data["genre"]).split(",")
        if genre != "Undefined"
    ]))

    # Languages
    game.languages = _class(soup, "languages")

    if game.languages:
        game.languages = game.languages.split(", ")
        game.languages.sort()
    else:
        game.languages = []

    # Players
    try:
        game.players = int(re.sub(r"[^\d]*", "", _class(soup, "num-of-players")))
    except (ValueError, TypeError):
        game.players = 0

    # Release date
    try:
        release_date = _itemprop(soup, "releaseDate")
        game.release_date = datetime.strptime(release_date, '%b %d, %Y')
    except ValueError:
        pass

    # Game size
    rom_size = _itemprop(soup, "romSize")

    if rom_size:
        value, unit = rom_size.split(" ")

        if unit.lower() == "blocks":
            value = int(value) // 8
        else:
            value = round(float(value) * (1024 if unit == "GB" else 1))
    else:
        value = None

    game.megabytes = value

    # Features
    game.amiibo = None  # unsupported
    game.demo = _aria_label(soup, "Download game demo opens in another window.") is not None
    game.dlc = _class(soup, "dlc", tag="section") is not None
    game.iaps = None  # unsupported
    game.free_to_play = data["msrp"] == '0'

    if game.platform == N3DS:
        game.street_pass = "StreetPass" in game.description
        game.virtual_console = soup.find("img", attrs={"alt": "Virtual Console"}) is not None

    if game.platform == SWITCH:
        game.local_multiplayer = None  # unsupported
        game.game_vouchers = _aria_label(soup, "Eligible for Game Vouchers") is not None
        game.nso_required = _aria_label(soup, "online-play") is not None
        game.save_data_cloud = _aria_label(soup, "save-data-cloud") is not None

    return game


@validate.nsuid
def game_info(*, nsuid: str) -> Game:
    """
        Given a valid nsuid for the NA region, it will retrieve the
    information of the game with that nsuid from Nintendo of America.

    Game data
    ---------
        * platform: str ["Nintendo 3DS", "Nintendo Switch"]
        * region: str ["NA"]
        * title: str
        * nsuid: str
        * product_code: str

        * slug: str

        * amiibo: bool (unsupported)
        * demo: bool
        * description: str
        * developer: str (optional)
        * dlc: bool
        * free_to_play: bool
        * genres: List[str]
        * iaps: bool (unsupported)
        * languages: List[str]
        * megabytes: int
        * players: int
        * publisher: str (optional)
        * release_date: datetime

        # 3DS Features
        * street_pass: bool
        * virtual_console: bool

        # Switch Features
        * local_multiplayer: bool (unsupported)
        * game_vouchers: bool
        * nso_required: bool
        * save_data_cloud: bool

    Parameters
    ----------
    nsuid: str
        Valid nsuid of a nintendo game.

    Returns
    -------
    nintendeals.classes.N3DSGame:
        3DS game from Nintendo of America.
    nintendeals.classes.SwitchGame:
        Switch game from Nintendo of America.
    None:
        No game with the provided nsuid was found on Nintendo of America.

    Raises
    -------
    nintendeals.exceptions.InvalidNsuidFormat
        The nsuid was either none or has an invalid format.
    """
    slug = algolia.find_by_nsuid(nsuid)

    log.info("Fetching info for %s", nsuid)

    if nsuid.startswith("5"):
        return _scrap(N3dsGame, slug=slug)

    if nsuid.startswith("7"):
        return _scrap(SwitchGame, slug=slug)

    return None



import json
import logging
from datetime import datetime
from typing import Union, Type

import requests
from bs4 import BeautifulSoup

from nintendeals import validate
from nintendeals.classes import N3dsGame, SwitchGame
from nintendeals.constants import JP

log = logging.getLogger(__name__)


def _get_extra_info(*, nsuid: str) -> json:
    response = requests.get(
        url="https://search.nintendo.jp/nintendo_soft/search.json",
        params={"q": nsuid}
    )

    if response.status_code != 200:
        return None

    items = response.json()["result"]["items"]

    if not items:
        return None

    return items[-1]


def _scrap_3ds(nsuid: str) -> N3dsGame:
    extra = _get_extra_info(nsuid=nsuid)

    if not extra:
        return None

    product_code = f"{extra['hard'][2:]}{extra['icode']}"

    game = N3dsGame(
        title=extra["title"],
        region=JP,
        nsuid=nsuid,
        product_code=product_code,
    )

    game.developer = extra.get("maker")
    game.description = extra.get("text")

    # Genres
    game.genres = list(sorted(extra.get("genre", [])))

    # Players
    try:
        game.players = max((
            int(p) for p in extra.get("player", [])
        ))
    except ValueError:
        game.players = 0

    # Release date
    try:
        game.release_date = datetime.strptime(
            extra["sdate"],
            '%Y.%m.%d'
        )
    except ValueError:
        pass

    # Features
    game.amiibo = extra.get("amiibo", "0") == "1"
    game.free_to_play = extra.get("dprice") == 0.0

    return game


def _scrap_switch(nsuid: str) -> SwitchGame:
    extra = _get_extra_info(nsuid=nsuid)

    if not extra:
        return None

    url = f"https://ec.nintendo.com/JP/jp/titles/{nsuid}"
    response = requests.get(url, allow_redirects=True)

    if response.status_code != 200:
        return None

    soup = BeautifulSoup(response.text, features="html.parser")

    script = next((
        s for s in soup.find_all("script")
        if "var NXSTORE = NXSTORE || {};" in str(s)
    ))

    json_data = next((
        line for line in str(script).split("\n")
        if "NXSTORE.titleDetail.jsonData = " in line
    ))

    data = json.loads(
        json_data.replace("NXSTORE.titleDetail.jsonData = ", "")[:-1]
    )

    product_code = f"{extra['hard'][2:]}{extra['icode']}"

    game = SwitchGame(
        region=JP,
        title=data["formal_name"],
        nsuid=nsuid,
        product_code=product_code,
    )

    game.developer = extra["maker"]
    game.description = data["description"]
    game.publisher = data["publisher"]["name"]

    # Genres
    game.genres = list(sorted(
        data.get("genre", "").split(" / ")
    ))

    # Languages
    game.languages = list(sorted(map(
        lambda lang: lang["name"], data.get("languages", [])
    )))

    # Players
    try:
        game.players = max(data.get("player_number", {}).values())
    except ValueError:
        game.players = 0

    # Release date
    try:
        game.release_date = datetime.strptime(
            data.get("release_date_on_eshop", ""),
            '%Y-%m-%d'
        )
    except ValueError:
        pass

    # Game size (in MBs)
    game.megabytes = round(data.get("total_rom_size", 0) / 1024 / 1024)

    # Other properties
    features = list(map(
        lambda feature: feature["name"], data.get("features", [])
    ))

    # Features
    game.amiibo = extra.get("amiibo", "0") == "1"
    game.demo = len(data.get("demos", [])) > 0
    game.dlc = data.get("has_aoc", False)
    game.free_to_play = extra.get("dprice") == 0.0
    game.iaps = data.get("in_app_purchase", False)

    game.game_vouchers = len(data.get("included_pretickets", [])) > 0
    game.local_multiplayer = data["player_number"].get("local_min", 0) > 0
    game.nso_required = "Nintendo Switch Online" in features
    game.save_data_cloud = data.get("cloud_backup_type") == "supported"

    return game


@validate.nsuid
def game_info(*, nsuid: str) -> Union[N3dsGame, SwitchGame, Type[None]]:
    """
        Given a valid nsuid for the JP region, it will retrieve the
    information of the game with that nsuid from Nintendo of Japan.

    Game data
    ---------
        * platform: str ["Nintendo 3DS", "Nintendo Switch"]
        * region: str ["JP"]
        * title: str
        * nsuid: str
        * product_code: str

        * demo: bool
        * description: str
        * developer: str
        * dlc: bool
        * free_to_play: bool
        * genres: List[str]
        * iaps: bool
        * languages: List[str]
        * local_multiplayer: bool
        * megabytes: int
        * publisher: str
        * release_date: datetime

        # Switch Features
        * save_data_cloud: bool
        * nso_required: bool

    Parameters
    ----------
    nsuid: str
        Valid nsuid of a nintendo game.

    Returns
    -------
    nintendeals.classes.N3DSGame:
        3DS game from Nintendo of Japan.
    nintendeals.classes.SwitchGame:
        Switch game from Nintendo of Japan.
    None:
        No game with the provided nsuid was found on Nintendo of Japan.

    Raises
    -------
    nintendeals.exceptions.InvalidNsuidFormat
        The nsuid was either none or has an invalid format.
    """
    if nsuid.startswith("5"):
        log.info("Fetching info for %s", nsuid)
        return _scrap_3ds(nsuid=nsuid)

    if nsuid.startswith("7"):
        log.info("Fetching info for %s", nsuid)
        return _scrap_switch(nsuid=nsuid)

    return None



from typing import Optional

from nintendeals import validate
from nintendeals.classes.games import Game
from nintendeals.constants import NA, EU, JP


class N3dsGame(Game):

    platform = "Nintendo 3DS"

    def __init__(
        self,
        region: str,
        title: str,
        nsuid: str = None,
        product_code: str = None,
    ):
        super().__init__(
            region=region,
            title=title,
            nsuid=nsuid,
            product_code=product_code
        )

        self.download_play: bool = None
        self.internet: bool = None
        self.motion_control: bool = None
        self.spot_pass: bool = None
        self.street_pass: bool = None
        self.virtual_console: bool = None

    @property
    def unique_id(self) -> Optional[str]:
        if not self.product_code:
            return None

        return self.product_code[-4:-1]

    @validate.country
    def url(self, *, country: str, lang: str = "en") -> Optional[str]:
        """
            Given a valid `country` code and an optional language it will
        provide a url (using the game's nsuid) that will redirect to the
        eShop page for this game.

        Parameters
        ----------
        country: str
            Valid alpha-2 code of the country.
        lang: str = "en"
            Valid iso-code for language.

        Returns
        -------
        str
            URL for the eShop of the game.

        Raises
        -------
        nintendeals.exceptions.InvalidAlpha2Code
            The `country` wasn't a valid alpha-2 code.
        """

        if self.region == NA and self.slug:
            return f"https://www.nintendo.com/{lang}_{country}/games/detail/{self.slug}/"

        if self.region == EU and self.slug:
            if country == "GB":
                country = "CO.UK"

            if country == "ZA":
                country = "CO.ZA"

            return f"https://www.nintendo.{country.lower()}/{lang}{self.slug}"

        if self.region == JP:
            return f"https://www.nintendo.co.jp/titles/{self.nsuid}"

        return None



import logging
from datetime import datetime
from typing import Iterator, Union

from nintendeals.classes import N3dsGame, SwitchGame
from nintendeals.helpers import search_filter
from nintendeals.noa import list_3ds_games, list_switch_games

log = logging.getLogger(__name__)


def _search_games(
    listing,
    *,
    title: str = None,
    released_at: datetime = None,
    released_after: datetime = None,
    released_before: datetime = None,
) -> Iterator[Union[N3dsGame, SwitchGame]]:
    params = {}

    if title:
        params = {"query": title}

    for game in listing(**params):
        if search_filter(
                game, title,
                released_at, released_after, released_before
        ):
            yield game


def search_3ds_games(
    *,
    title: str = None,
    released_at: datetime = None,
    released_after: datetime = None,
    released_before: datetime = None,
) -> Iterator[N3dsGame]:
    """
        Search and filter all the 3DS games in Nintendo of America.

    Parameters
    ----------
    title: str (Optional)
        String that should be contained in the title of each game.
    released_at: datetime (Optional)
        Keep only those games that have been released on this date.
    released_after: datetime (Optional)
        Keep only those games that have been released after this date.
    released_before: datetime (Optional)
        Keep only those games that have been released before this date.

    Yields
    -------
    nintendeals.classes.N3dsGame:
        3DS game from Nintendo of America.

    See Also
    ---------
    nintendeals.noa.list_3ds_games
    """
    log.info("Searching Nintendo 3DS games")

    yield from _search_games(
        list_3ds_games,
        title=title,
        released_at=released_at,
        released_after=released_after,
        released_before=released_before,
    )


def search_switch_games(
    *,
    title: str = None,
    released_at: datetime = None,
    released_after: datetime = None,
    released_before: datetime = None,
) -> Iterator[SwitchGame]:
    """
        Search and filter all the Switch games in Nintendo of America.

    Parameters
    ----------
    title: str (Optional)
        String that should be contained in the title of each game.
    released_at: datetime (Optional)
        Keep only those games that have been released on this date.
    released_after: datetime (Optional)
        Keep only those games that have been released after this date.
    released_before: datetime (Optional)
        Keep only those games that have been released before this date.

    Yields
    -------
    nintendeals.classes.SwitchGame:
        Switch game from Nintendo of America.

    See Also
    ---------
    nintendeals.noa.list_switch_games
    """
    log.info("Searching Nintendo Switch games")

    yield from _search_games(
        list_switch_games,
        title=title,
        released_at=released_at,
        released_after=released_after,
        released_before=released_before,
    )
